"""
Defines utility functions used for data visualization.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
import sys
from typing import Optional
from typing import Union
from typing import Any
from typing import TYPE_CHECKING
import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go  # type: ignore
from .. import transformations
from . import graphics_common
from . import graphics_common_3d
from .preprocessing_3d import add_data__global_axes
from ..postprocessing.basic_forces import basic_forces
from ..ops import element
from ..model import Model

if TYPE_CHECKING:
    from ..solver import Analysis


nparr = npt.NDArray[np.float64]


def force_scaling_factor(ref_len, fmax, factor):
    """
    Applies a scaling factor to basic forces.

    Arguments:
      ref_len: :func:`~osmg.model.Model.reference_length` of the
        model.
      fmax: Largest value in the basic forces.
      factor: Required scaling factor.

    """
    if fmax == 0.00:
        result = 0.00
    else:
        result = ref_len / fmax * factor
    return result


def interp_3d_deformation(elm, u_i, r_i, u_j, r_j, num_points):
    """
    Given the deformations of the ends of a Bernoulli beam,
    use its shape functions to obtain intermediate points.

    Arguments:
      element: A line element
      u_i: 3 displacements at end i, global system
      r_i: 3 rotations at end i, global system
      u_j, r_j: similar to u_i, r_i.
      num_points: Number of interpolation points

    Returns:
      Displacements (global system) and rotations (local system). The
        rotations are needed for plotting the deformed shape with
        extruded frame elements.

    """

    x_vec: nparr = elm.geomtransf.x_axis
    y_vec: nparr = elm.geomtransf.y_axis
    z_vec: nparr = np.cross(x_vec, y_vec)

    # global -> local transformation matrix
    transf_global2local = transformations.transformation_matrix(
        x_vec, y_vec, z_vec
    )
    transf_local2global = transf_global2local.T

    u_i_global = u_i
    r_i_global = r_i
    u_j_global = u_j
    r_j_global = r_j

    u_i_local = transf_global2local @ u_i_global
    r_i_local = transf_global2local @ r_i_global
    u_j_local = transf_global2local @ u_j_global
    r_j_local = transf_global2local @ r_j_global

    # discrete sample location parameter
    t_vec = np.linspace(0.00, 1.00, num=num_points)
    p_i = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
    p_j = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
    len_clr = np.linalg.norm(p_i - p_j)

    # shape function matrices
    nx_mat = np.column_stack((1.0 - t_vec, t_vec))
    nyz_mat = np.column_stack(
        (
            1.0 - 3.0 * t_vec**2 + 2.0 * t_vec**3,
            (t_vec - 2.0 * t_vec**2 + t_vec**3) * len_clr,
            3.0 * t_vec**2 - 2.0 * t_vec**3,
            (-(t_vec**2) + t_vec**3) * len_clr,
        )
    )
    nyz_derivative_mat = np.column_stack(
        (
            -6.0 * t_vec + 6.0 * t_vec**2,
            (1 - 4.0 * t_vec + 3.0 * t_vec**2) * len_clr,
            6.0 * t_vec - 6.0 * t_vec**2,
            (-2.0 * t_vec + 3.0 * t_vec**2) * len_clr,
        )
    )

    # axial deformation
    d_x_local = nx_mat @ np.array([u_i_local[0], u_j_local[0]])

    # bending deformation along the local xy plane
    d_y_local = nyz_mat @ np.array(
        [u_i_local[1], r_i_local[2], u_j_local[1], r_j_local[2]]
    )

    # bending deformation along the local xz plane
    d_z_local = nyz_mat @ np.array(
        [u_i_local[2], -r_i_local[1], u_j_local[2], -r_j_local[1]]
    )

    # torsional deformation
    r_x_local = nx_mat @ np.array([r_i_local[0], r_j_local[0]])

    # bending rotation around the local z axis
    r_z_local = (
        nyz_derivative_mat
        @ np.array([u_i_local[1], r_i_local[2], u_j_local[1], r_j_local[2]])
        / len_clr
    )

    # bending rotation around the local y axis
    r_y_local = (
        nyz_derivative_mat
        @ np.array([-u_i_local[2], r_i_local[1], -u_j_local[2], r_j_local[1]])
        / len_clr
    )

    # all deformations
    d_local = np.column_stack((d_x_local, d_y_local, d_z_local))

    # all rotations
    r_local = np.column_stack((r_x_local, r_y_local, r_z_local))

    d_global = (transf_local2global @ d_local.T).T

    return d_global, r_local


def interp_3d_points(elm, d_global, num_points, scaling):
    """
    Calculates intermediate points based on end locations and
    deformations.

    """

    if not isinstance(elm, element.TrussBar):
        p_i = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        p_j = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
    else:
        p_i = np.array(elm.nodes[0].coords)
        p_j = np.array(elm.nodes[1].coords)
    element_point_samples: nparr = np.column_stack(
        (
            np.linspace(p_i[0], p_j[0], num=num_points),
            np.linspace(p_i[1], p_j[1], num=num_points),
            np.linspace(p_i[2], p_j[2], num=num_points),
        )
    )

    interpolation_points = element_point_samples + d_global * scaling

    return interpolation_points


def add_data__extruded_line_elms_deformed_mesh(
    analysis, case_name, data_dict, list_of_line_elems, step, scaling
):
    """
    Adds a trace containing frame element extrusion mesh
    in its deformed state.

    """

    if not list_of_line_elems:
        return
    x_list: list[float] = []
    y_list: list[float] = []
    z_list: list[float] = []
    i_list: list[int] = []
    j_list: list[int] = []
    k_list: list[int] = []
    intensity: list[float] = []
    index = 0
    for elm in list_of_line_elems:
        if elm.visibility.hidden_when_extruded:
            continue
        if isinstance(elm, element.TrussBar):
            if elm.outside_shape is None:
                continue
        else:
            if elm.section.outside_shape is None:
                continue

        num_points = 8
        # translations and rotations at the offset ends
        u_i = analysis.results[case_name].node_displacements[elm.nodes[0].uid][
            step
        ][0:3]
        r_i = analysis.results[case_name].node_displacements[elm.nodes[0].uid][
            step
        ][3:6]
        u_j = analysis.results[case_name].node_displacements[elm.nodes[1].uid][
            step
        ][0:3]
        r_j = analysis.results[case_name].node_displacements[elm.nodes[1].uid][
            step
        ][3:6]
        # transferring them to the clear element ends
        if isinstance(elm, element.TrussBar):
            offset_i = np.zeros(3)
            offset_j = np.zeros(3)
            (
                x_vec,
                y_vec,
                z_vec,
            ) = transformations.local_axes_from_points_and_angle(
                np.array(elm.nodes[0].coords),
                np.array(elm.nodes[1].coords),
                0.00,
            )
            outside_shape = elm.outside_shape
        else:
            offset_i = elm.geomtransf.offset_i
            offset_j = elm.geomtransf.offset_j
            x_vec = elm.geomtransf.x_axis
            y_vec = elm.geomtransf.y_axis
            z_vec = elm.geomtransf.z_axis
            outside_shape = elm.section.outside_shape
        u_i_o = transformations.offset_transformation(offset_i, u_i, r_i)
        u_j_o = transformations.offset_transformation(offset_j, u_j, r_j)
        if isinstance(elm, element.TrussBar):
            d_global = np.column_stack(
                (
                    np.linspace(u_i[0], u_j[0], num_points),
                    np.linspace(u_i[1], u_j[1], num_points),
                    np.linspace(u_i[2], u_j[2], num_points),
                )
            )
            r_local = np.zeros((num_points, 3))
            interpolation_points = interp_3d_points(
                elm, d_global, num_points, scaling
            )
        else:
            d_global, r_local = interp_3d_deformation(
                elm, u_i_o, r_i, u_j_o, r_j, num_points
            )
            interpolation_points = interp_3d_points(
                elm, d_global, num_points, scaling
            )
        for i in range(num_points - 1):
            loc_i_global = interpolation_points[i, :]
            loc_j_global = interpolation_points[i + 1, :]
            rot_i_local = r_local[i, :]
            rot_j_local = r_local[i + 1, :]

            assert outside_shape is not None
            loop = outside_shape.halfedges
            for halfedge in loop:
                assert halfedge.nxt is not None
                z_a = halfedge.vertex.coords[0]
                y_a = halfedge.vertex.coords[1]
                z_b = halfedge.nxt.vertex.coords[0]
                y_b = halfedge.nxt.vertex.coords[1]
                defo_ia_global = (
                    z_a * z_vec
                    + y_a * y_vec
                    + (
                        -rot_i_local[2] * y_a * x_vec
                        + rot_i_local[1] * z_a * x_vec
                        + rot_i_local[0] * y_a * z_vec
                        - rot_i_local[0] * z_a * y_vec
                    )
                    * scaling
                )
                defo_ja_global = (
                    z_a * z_vec
                    + y_a * y_vec
                    + (
                        -rot_j_local[2] * y_a * x_vec
                        + rot_j_local[1] * z_a * x_vec
                        + rot_j_local[0] * y_a * z_vec
                        - rot_j_local[0] * z_a * y_vec
                    )
                    * scaling
                )
                defo_ib_global = (
                    z_b * z_vec
                    + y_b * y_vec
                    + (
                        -rot_i_local[2] * y_b * x_vec
                        + rot_i_local[1] * z_b * x_vec
                        + rot_i_local[0] * y_b * z_vec
                        - rot_i_local[0] * z_b * y_vec
                    )
                    * scaling
                )
                defo_jb_global = (
                    z_b * z_vec
                    + y_b * y_vec
                    + (
                        -rot_j_local[2] * y_b * x_vec
                        + rot_i_local[1] * z_b * x_vec
                        + rot_j_local[0] * y_b * z_vec
                        - rot_j_local[0] * z_b * y_vec
                    )
                    * scaling
                )
                loc0 = loc_i_global + defo_ia_global
                loc1 = loc_j_global + defo_ja_global
                loc2 = loc_j_global + defo_jb_global
                loc3 = loc_i_global + defo_ib_global
                x_list.extend((loc0[0], loc1[0], loc2[0], loc3[0]))
                y_list.extend((loc0[1], loc1[1], loc2[1], loc3[1]))
                z_list.extend((loc0[2], loc1[2], loc2[2], loc3[2]))
                i_list.extend((index + 0, index + 0))
                j_list.extend((index + 1, index + 2))
                k_list.extend((index + 2, index + 3))
                intensity.append(
                    float(np.sqrt(d_global[i, :] @ d_global[i, :]))
                )
                intensity.append(
                    float(np.sqrt(d_global[i + 1, :] @ d_global[i + 1, :]))
                )
                intensity.append(
                    float(np.sqrt(d_global[i + 1, :] @ d_global[i + 1, :]))
                )
                intensity.append(
                    float(np.sqrt(d_global[i, :] @ d_global[i, :]))
                )
                index += 4
    data_dict.append(
        {
            "type": "mesh3d",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "colorscale": [[0, "blue"], [1.0, "red"]],
            "i": i_list,
            "j": j_list,
            "k": k_list,
            "intensity": intensity,
            "colorbar": {"title": "Displacement", "ypad": 300},
            "hoverinfo": "skip",
            "opacity": 0.65,
        }
    )


def add_data__line_elms_deformed(
    analysis, case_name, data_dict, list_of_line_elems, step, scaling
):
    """
    Adds a trace containing frame element centroidal axis lines
    in their deformed state.

    """

    if not list_of_line_elems:
        return
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []

    for elm in list_of_line_elems:
        if elm.visibility.hidden_at_line_plots:
            continue
        num_points = 20
        u_i = analysis.results[case_name].node_displacements[elm.nodes[0].uid][
            step
        ][0:3]
        r_i = analysis.results[case_name].node_displacements[elm.nodes[0].uid][
            step
        ][3:6]
        u_j = analysis.results[case_name].node_displacements[elm.nodes[1].uid][
            step
        ][0:3]
        r_j = analysis.results[case_name].node_displacements[elm.nodes[1].uid][
            step
        ][3:6]
        if not isinstance(elm, element.TrussBar):
            # transferring them to the clear element ends
            offset_i = elm.geomtransf.offset_i
            offset_j = elm.geomtransf.offset_j
            u_i_o = transformations.offset_transformation(offset_i, u_i, r_i)
            u_j_o = transformations.offset_transformation(offset_j, u_j, r_j)
            d_global, _ = interp_3d_deformation(
                elm, u_i_o, r_i, u_j_o, r_j, num_points
            )
        else:
            # for a truss member, just connect the two ends
            d_global = np.column_stack(
                (
                    np.linspace(u_i[0], u_j[0], num_points),
                    np.linspace(u_i[1], u_j[1], num_points),
                    np.linspace(u_i[2], u_j[2], num_points),
                )
            )
        interpolation_points = interp_3d_points(
            elm, d_global, num_points, scaling
        )
        for i in range(len(interpolation_points) - 1):
            x_list.extend(
                (
                    interpolation_points[i, 0],
                    interpolation_points[i + 1, 0],
                    None,
                )
            )
            y_list.extend(
                (
                    interpolation_points[i, 1],
                    interpolation_points[i + 1, 1],
                    None,
                )
            )
            z_list.extend(
                (
                    interpolation_points[i, 2],
                    interpolation_points[i + 1, 2],
                    None,
                )
            )

    data_dict.append(
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "hoverinfo": "skip",
            "line": {"width": 5, "color": graphics_common.FRAME_COLOR},
        }
    )


def add_data__line_elm_offsets_deformed(
    analysis, case_name, data_dict, list_of_line_elems, step, scaling
):
    """
    Adds a trace containing frame element rigid offset lines
    in their deformed state.

    """

    if not list_of_line_elems:
        return
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    for elm in list_of_line_elems:
        if isinstance(elm, element.TrussBar):
            continue
        if np.array_equal(elm.geomtransf.offset_i, np.zeros(3)):
            if np.array_equal(elm.geomtransf.offset_j, np.zeros(3)):
                continue
        p_i: nparr = np.array(elm.nodes[0].coords)
        p_io: nparr = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        offset_i = elm.geomtransf.offset_i
        u_i: nparr = np.array(
            analysis.results[case_name].node_displacements[elm.nodes[0].uid][
                step
            ][0:3]
        )
        r_i: nparr = np.array(
            analysis.results[case_name].node_displacements[elm.nodes[0].uid][
                step
            ][3:6]
        )
        u_io: nparr = transformations.offset_transformation(offset_i, u_i, r_i)

        p_j: nparr = np.array(elm.nodes[1].coords)
        p_jo: nparr = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
        offset_j = elm.geomtransf.offset_j
        u_j: nparr = np.array(
            analysis.results[case_name].node_displacements[elm.nodes[1].uid][
                step
            ][0:3]
        )
        r_j: nparr = np.array(
            analysis.results[case_name].node_displacements[elm.nodes[1].uid][
                step
            ][3:6]
        )
        u_jo: nparr = transformations.offset_transformation(offset_j, u_j, r_j)

        x_i = p_i + u_i * scaling
        x_io = p_io + u_io * scaling
        x_j = p_j + u_j * scaling
        x_jo = p_jo + u_jo * scaling

        x_list.extend((x_i[0], x_io[0], None))
        y_list.extend((x_i[1], x_io[1], None))
        z_list.extend((x_i[2], x_io[2], None))
        x_list.extend((x_j[0], x_jo[0], None))
        y_list.extend((x_j[1], x_jo[1], None))
        z_list.extend((x_j[2], x_jo[2], None))

    data_dict.append(
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "hoverinfo": "skip",
            "line": {"width": 8, "color": graphics_common.OFFSET_COLOR},
        }
    )


def add_data__frames_undeformed(data_dict, list_of_line_elems):
    """
    Adds a trace containing frame element centroidal axis lines

    """

    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []

    for elm in list_of_line_elems:
        if isinstance(elm, element.TrussBar):
            p_i = np.array(elm.nodes[0].coords)
            p_j = np.array(elm.nodes[1].coords)
        else:
            p_i = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
            p_j = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j

        x_list.extend((p_i[0], p_j[0], None))
        y_list.extend((p_i[1], p_j[1], None))
        z_list.extend((p_i[2], p_j[2], None))
    data_dict.append(
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "hoverinfo": "skip",
            "line": {"width": 5, "color": graphics_common.BEAM_MESH_COLOR},
        }
    )


def add_data__nodes_deformed(
    analysis, case_name, data_dict, list_of_nodes, step, scaling, function
):
    """
    Adds a trace containing nodes in their deformed locations.

    """
    ids_list = [int(node.uid) for node in list_of_nodes]
    location_data = np.full((len(list_of_nodes), 3), 0.00)
    displacement_data = np.full((len(list_of_nodes), 6), 0.00)
    for i, node in enumerate(list_of_nodes):
        location_data[i, :] = node.coords
        displacement_data[i, :] = analysis.results[
            case_name
        ].node_displacements[node.uid][step]
    dist = np.sqrt(
        displacement_data[:, 0] ** 2
        + displacement_data[:, 1] ** 2
        + displacement_data[:, 2] ** 2
    )
    dist = np.reshape(dist, (-1, 1))
    ids = np.reshape(np.array(ids_list), (-1, 1))
    displacement_data = np.concatenate((displacement_data, dist, ids), 1)
    restraint_symbols = []
    for node in list_of_nodes:
        if True in node.restraint:
            restraint_symbols.append("fixed")
        else:
            restraint_symbols.append(function)

    marker = [
        graphics_common_3d.node_marker[sym][0] for sym in restraint_symbols
    ]
    size = [
        graphics_common_3d.node_marker[sym][1] for sym in restraint_symbols
    ]
    if function == "internal":
        color = graphics_common.NODE_INTERNAL_COLOR
    else:
        color = graphics_common.NODE_PRIMARY_COLOR

    data_dict.append(
        {
            "type": "scatter3d",
            "mode": "markers",
            "x": location_data[:, 0] + displacement_data[:, 0] * scaling,
            "y": location_data[:, 1] + displacement_data[:, 1] * scaling,
            "z": location_data[:, 2] + displacement_data[:, 2] * scaling,
            "customdata": displacement_data,
            "hovertemplate": (
                "ux: %{customdata[0]:.6g}<br>"
                + "uy: %{customdata[1]:.6g}<br>"
                + "uz: %{customdata[2]:.6g}<br>"
                + "combined: %{customdata[6]:.6g}<br>"
                + "rx: %{customdata[3]:.6g} (rad)<br>"
                + "ry: %{customdata[4]:.6g} (rad)<br>"
                + "rz: %{customdata[5]:.6g} (rad)<br>"
                + "<extra>Node %{customdata[7]:d}</extra>"
            ),
            "marker": {
                "symbol": marker,
                "color": color,
                "size": size,
                "line": {"color": color, "width": 4},
            },
        }
    )


def get_auto_scaling_deformation(analysis, case_name, mdl, step):
    """
    Automatically calculate a scaling value that makes the maximum
    displacement appear approximately 10% of the largest dimention of
    the building's bounding box.

    """

    ref_len = mdl.reference_length()
    # maximum displacement
    max_d = 0.00
    elms: list[Union[element.ElasticBeamColumn, element.DispBeamColumn]] = []
    elms.extend(mdl.list_of_specific_element(element.ElasticBeamColumn))
    elms.extend(mdl.list_of_specific_element(element.DispBeamColumn))

    for elm in elms:
        u_i = analysis.results[case_name].node_displacements[elm.nodes[0].uid][
            step
        ][0:3]
        r_i = analysis.results[case_name].node_displacements[elm.nodes[0].uid][
            step
        ][3:6]
        u_j = analysis.results[case_name].node_displacements[elm.nodes[1].uid][
            step
        ][0:3]
        r_j = analysis.results[case_name].node_displacements[elm.nodes[1].uid][
            step
        ][3:6]
        d_global, _ = interp_3d_deformation(elm, u_i, r_i, u_j, r_j, 3)
        max_d = np.maximum(max_d, np.max(np.abs(d_global)))
    # scaling factor: max_d scaled = 10% of the reference length
    if max_d > 1.00e-14:
        scaling = ref_len / max_d * 0.1
    else:
        # no infinite scaling, thank you
        scaling = 1.00

    # never scale things down
    # (usually when this is required, things have gone bad
    #  and we should be able to realize that immediately)
    scaling = max(scaling, 1.00)
    return scaling


def show_deformed_shape(
    analysis: Analysis,
    case_name: str,
    step: int,
    scaling: float,
    extrude: bool,
    camera: Optional[dict[str, object]] = None,
    subset_model: Model = None,
    animation: bool = False,
    init_step: int = 0,
    step_skip: int = 0,
    to_html_file: Optional[str] = None,
) -> dict[str, Any]:
    """
    Visualize the model in its deformed state

    Arguments:
      analysis: an analysis object
      case_name: the name of the load_case to be visualized
      step: the analysis step to be visualized
      scaling: scaling factor for the deformations. If 0.00 is
        provided, the scaling factor is calculated automatically.
      extrude: wether to extrude frame elements
      camera: custom positioning of the camera
      subset_model: subset model used to only show certain
        components
      animation: show all frames up to the one identified with
        `step`
      init_step: starting step, in case of animation
      step_skip: how many frames to skip to reduce the number of
        frames in case an animation.
      to_html_file: If a path is specified, the figure is written in
        an html file instead of being shown.

    """

    if subset_model:
        # if a subset model is specified, only show its components
        mdl = subset_model
    else:
        # otherwise show the entire model
        mdl = analysis.mdl

    # calculate a nice scaling factor if 0.00 is passed
    if scaling == 0.00:
        scaling = get_auto_scaling_deformation(analysis, case_name, mdl, step)

    # instantiate layout and datastructures
    layout = graphics_common_3d.global_layout(mdl, camera)
    data_dict: list[dict[str, object]] = []
    frame_data_dict: list[list[dict[str, object]]] = []

    # gather lists of associated objects
    list_of_line_elems: list[element.Element] = []
    list_of_line_elems.extend(mdl.list_of_specific_element(element.TrussBar))
    list_of_line_elems.extend(
        mdl.list_of_specific_element(element.ElasticBeamColumn)
    )
    list_of_line_elems.extend(
        mdl.list_of_specific_element(element.DispBeamColumn)
    )
    list_of_primary_nodes = mdl.list_of_primary_nodes()
    list_of_internal_nodes = mdl.list_of_internal_nodes()
    # list_of_parent_nodes = mdl.list_of_parent_nodes()

    # add data for the global axes
    ref_len = mdl.reference_length()

    add_data__global_axes(data_dict, ref_len)
    if animation:
        first_step = init_step
    else:
        first_step = step

    add_data__nodes_deformed(
        analysis,
        case_name,
        data_dict,
        list_of_primary_nodes,
        first_step,
        scaling,
        "free",
    )
    add_data__nodes_deformed(
        analysis,
        case_name,
        data_dict,
        list_of_internal_nodes,
        first_step,
        scaling,
        "internal",
    )
    # add data for the frame elements as lines
    add_data__line_elm_offsets_deformed(
        analysis, case_name, data_dict, list_of_line_elems, first_step, scaling
    )
    add_data__line_elms_deformed(
        analysis, case_name, data_dict, list_of_line_elems, first_step, scaling
    )
    # add data for the extruded frame elements
    if extrude:
        add_data__extruded_line_elms_deformed_mesh(
            analysis,
            case_name,
            data_dict,
            list_of_line_elems,
            first_step,
            scaling,
        )

    # create the plot

    def frame_args(duration):
        return {
            "frame": {"duration": duration},
            "mode": "immediate",
            "fromcurrent": True,
            "transition": {"duration": duration, "easing": "linear"},
        }

    if animation:
        step_of_frame = []
        for j in range(first_step, step + 1, step_skip + 1):
            step_of_frame.append(j)
        for _, j in enumerate(step_of_frame):
            frame_data_dict.append([])
            add_data__global_axes(frame_data_dict[-1], ref_len)
            # add_data__global_axes(frame_data_dict[0], ref_len)
            add_data__nodes_deformed(
                analysis,
                case_name,
                frame_data_dict[-1],
                list_of_primary_nodes,
                j,
                scaling,
                "free",
            )
            add_data__nodes_deformed(
                analysis,
                case_name,
                frame_data_dict[-1],
                list_of_internal_nodes,
                j,
                scaling,
                "internal",
            )
            # add data for the frame elements as lines
            add_data__line_elm_offsets_deformed(
                analysis,
                case_name,
                frame_data_dict[-1],
                list_of_line_elems,
                j,
                scaling,
            )
            add_data__line_elms_deformed(
                analysis,
                case_name,
                frame_data_dict[-1],
                list_of_line_elems,
                j,
                scaling,
            )
            # add data for the extruded frame elements
            if extrude:
                add_data__extruded_line_elms_deformed_mesh(
                    analysis,
                    case_name,
                    frame_data_dict[-1],
                    list_of_line_elems,
                    j,
                    scaling,
                )

        fig_datastructure = dict(
            data=data_dict,
            layout=layout,
            frames=[
                go.Frame(data=frame_data_dict[j], name=str(step_of_frame[j]))
                for j in range(len(step_of_frame))
            ],
        )

        fig = go.Figure(fig_datastructure)
        sliders = [
            {
                "pad": {"b": 10, "t": 60},
                "len": 0.9,
                "x": 0.1,
                "y": 0,
                "steps": [
                    {
                        "args": [[f.name], frame_args(0)],
                        "label": str(k),
                        "method": "animate",
                    }
                    for k, f in enumerate(fig.frames)
                ],
            }
        ]
        # Layout
        fig.update_layout(
            updatemenus=[
                {
                    "buttons": [
                        {
                            "args": [None, frame_args(50)],
                            "label": "&#9654;",  # play symbol
                            "method": "animate",
                        },
                        {
                            "args": [[None], frame_args(0)],
                            "label": "&#9724;",  # pause symbol
                            "method": "animate",
                        },
                    ],
                    "direction": "left",
                    "pad": {"r": 10, "t": 70},
                    "type": "buttons",
                    "x": 0.1,
                    "y": 0,
                }
            ],
            sliders=sliders,
        )
    else:  # (not animation)
        fig_datastructure = dict(data=data_dict, layout=layout)
        fig = go.Figure(fig_datastructure)

    # show the plot (if it's not a test)
    if "pytest" not in sys.modules:
        if to_html_file:
            fig.write_html(to_html_file)
        else:
            fig.show()

    # return plot-related metadata
    metadata = {"scaling": scaling}
    return metadata


def show_basic_forces(
    analysis,
    case_name,
    step,
    scaling_global,
    scaling_n,
    scaling_q,
    scaling_m,
    scaling_t,
    num_points,
    force_conversion=1.00,
    moment_conversion=1.00,
    global_axes=False,
    camera=None,
    subset_model=None,
    to_html_file=None,
):
    """
    Visualize the model and plot the frame element basic forces.

    Arguments:
      analysis: an analysis object
      case_name: the name of the load_case to be visualized
      step: the analysis step to be visualized
      scaling_global: I don't even remember what this
        does. It's kind of a mess right now.
      scaling_n:
      scaling_q:
      scaling_m:
      scaling_t:
      num_points: number of points to include in the basic force
        curves
      force_conversion: Conversion factor to be applied at the
        hover box data for forces (for unit conversions)
      moment_conversion: Conversion factor to be applied at the
        hover box data for moments (for unit conversions)
      global_axes: whether to show global axes
      camera: custom positioning of the camera
      subset_model: use this model instead of the one contained in the
        analysis object. It needs to be a subset of the original
        model. This can be used to only show the results for some part
        of a large model.
      to_html_file: If a path is specified, the figure is written in
        an html file instead of being shown.

    """

    if subset_model:
        mdl = subset_model
    else:
        mdl = analysis.mdl

    elms: list[
        Union[
            element.TrussBar, element.ElasticBeamColumn, element.DispBeamColumn
        ]
    ] = []
    elms.extend(mdl.list_of_specific_element(element.TrussBar))
    elms.extend(mdl.list_of_specific_element(element.ElasticBeamColumn))
    elms.extend(mdl.list_of_specific_element(element.DispBeamColumn))

    list_of_line_elements = [
        elm for elm in elms if not elm.visibility.skip_opensees_definition
    ]

    layout = graphics_common_3d.global_layout(mdl, camera)
    data_dict: list[dict[str, object]] = []

    # draw the frames
    add_data__frames_undeformed(data_dict, list_of_line_elements)
    # we also add axes so that we can see 2D plots
    ref_len = mdl.reference_length()
    if global_axes:
        add_data__global_axes(data_dict, ref_len)

    # Plot options:
    # a: axial
    # b: shear in local Y and Z
    # c: moment in Y and Z
    # d: torsion
    # e: shear combined
    # f: moment combined
    x1_a: list[Optional[float]] = []
    y1_a: list[Optional[float]] = []
    z1_a: list[Optional[float]] = []
    colors1_a: list[Optional[str]] = []
    customdata_a: list[Optional[float]] = []

    x1_b: list[Optional[float]] = []
    y1_b: list[Optional[float]] = []
    z1_b: list[Optional[float]] = []
    colors1_b: list[Optional[str]] = []
    customdata_b: list[Optional[float]] = []

    x1_c: list[Optional[float]] = []
    y1_c: list[Optional[float]] = []
    z1_c: list[Optional[float]] = []
    colors1_c: list[Optional[str]] = []
    customdata_c: list[Optional[float]] = []

    x1_d: list[Optional[float]] = []
    y1_d: list[Optional[float]] = []
    z1_d: list[Optional[float]] = []
    colors1_d: list[Optional[str]] = []
    customdata_d: list[Optional[float]] = []

    x1_e: list[Optional[float]] = []
    y1_e: list[Optional[float]] = []
    z1_e: list[Optional[float]] = []
    colors1_e: list[Optional[str]] = []
    customdata_e: list[Optional[float]] = []

    x1_f: list[Optional[float]] = []
    y1_f: list[Optional[float]] = []
    z1_f: list[Optional[float]] = []
    colors1_f: list[Optional[str]] = []
    customdata_f: list[Optional[float]] = []

    # (we do this to determine the internal forces
    #  for all elements before we even start plotting
    #  them, to be able to compute a nice scaling factor
    #  without having to then recalculate the basic forces)
    nx_vecs = {}
    qy_vecs = {}
    qz_vecs = {}
    tx_vecs = {}
    mz_vecs = {}
    my_vecs = {}

    for elm in list_of_line_elements:
        if elm.visibility.hidden_basic_forces:
            continue

        forces = basic_forces(
            analysis, case_name, step, elm, num_points, as_tuple=True
        )
        assert isinstance(forces, tuple)
        nx_vec, qy_vec, qz_vec, tx_vec, mz_vec, my_vec = forces
        assert isinstance(nx_vec, np.ndarray)
        assert isinstance(qy_vec, np.ndarray)
        assert isinstance(qz_vec, np.ndarray)
        assert isinstance(tx_vec, np.ndarray)
        assert isinstance(mz_vec, np.ndarray)
        assert isinstance(my_vec, np.ndarray)

        # store results in the preallocated arrays

        nx_vecs[elm.uid] = nx_vec * force_conversion
        qy_vecs[elm.uid] = qy_vec * force_conversion
        qz_vecs[elm.uid] = qz_vec * force_conversion
        tx_vecs[elm.uid] = tx_vec * moment_conversion
        my_vecs[elm.uid] = my_vec * moment_conversion
        mz_vecs[elm.uid] = mz_vec * moment_conversion

    # calculate scaling factors
    ref_len = mdl.reference_length()
    factor = 0.05
    nx_max = np.max(np.abs(np.column_stack(list(nx_vecs.values()))))
    scaling_n = force_scaling_factor(ref_len, nx_max, factor)
    if scaling_n > 1.0e8:
        scaling_t = 1.00
    qy_max = np.max(np.abs(np.column_stack(list(qy_vecs.values()))))
    qz_max = np.max(np.abs(np.column_stack(list(qz_vecs.values()))))
    scaling_qy = force_scaling_factor(ref_len, qy_max, factor)
    scaling_qz = force_scaling_factor(ref_len, qz_max, factor)
    if scaling_qy > 0.00 and scaling_qz > 0.00:
        scaling_q = np.minimum(scaling_qy, scaling_qz)
    elif scaling_qy == 0.00:
        scaling_q = scaling_qz
    elif scaling_qz == 0.00:
        scaling_q = scaling_qy
    else:
        scaling_q = 0.00
    if scaling_q > 1.0e8:
        scaling_q = 1.00
    my_max = np.max(np.abs(np.column_stack(list(my_vecs.values()))))
    mz_max = np.max(np.abs(np.column_stack(list(mz_vecs.values()))))
    scaling_my = force_scaling_factor(ref_len, my_max, factor)
    scaling_mz = force_scaling_factor(ref_len, mz_max, factor)
    if scaling_my > 0.00 and scaling_mz > 0.00:
        scaling_m = np.minimum(scaling_my, scaling_mz)
    elif scaling_my == 0.00:
        scaling_m = scaling_mz
    elif scaling_mz == 0.00:
        scaling_m = scaling_my
    else:
        scaling_m = 0.00
    if scaling_m > 1.0e8:
        scaling_m = 1.00

    for elm in list_of_line_elements:
        if elm.visibility.hidden_basic_forces:
            continue

        # retrieve results from the preallocated arrays
        nx_vec = nx_vecs[elm.uid]
        qy_vec = qy_vecs[elm.uid]
        qz_vec = qz_vecs[elm.uid]
        tx_vec = tx_vecs[elm.uid]
        my_vec = my_vecs[elm.uid]
        mz_vec = mz_vecs[elm.uid]
        if isinstance(
            elm, (element.ElasticBeamColumn, element.DispBeamColumn)
        ):
            x_vec = elm.geomtransf.x_axis
            y_vec = elm.geomtransf.y_axis
            z_vec = elm.geomtransf.z_axis
            i_pos = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
            p_i = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
            p_j = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
        else:
            p_i = np.array(elm.nodes[0].coords)
            p_j = np.array(elm.nodes[1].coords)
            i_pos = np.array(elm.nodes[0].coords)
            (
                x_vec,
                y_vec,
                z_vec,
            ) = transformations.local_axes_from_points_and_angle(
                p_i, p_j, 0.00
            )

        len_clr = np.linalg.norm(p_i - p_j)
        t_vec = np.linspace(0.00, len_clr, num=num_points)

        for i in range(num_points - 1):
            p_start = i_pos + t_vec[i] * x_vec
            p_end = i_pos + t_vec[i + 1] * x_vec

            # axial load
            p_i = p_start + nx_vec[i] * y_vec * scaling_n * scaling_global
            p_j = p_end + nx_vec[i + 1] * y_vec * scaling_n * scaling_global

            x1_a.extend((p_i[0], p_j[0], None))
            y1_a.extend((p_i[1], p_j[1], None))
            z1_a.extend((p_i[2], p_j[2], None))
            customdata_a.extend((nx_vec[i], nx_vec[i + 1], None))
            colors1_a.extend(["red"] * 3)

            # torsion
            p_i = p_start + tx_vec[i] * z_vec * scaling_t * scaling_global
            p_j = p_end + tx_vec[i + 1] * z_vec * scaling_t * scaling_global
            x1_d.extend((p_i[0], p_j[0], None))
            y1_d.extend((p_i[1], p_j[1], None))
            z1_d.extend((p_i[2], p_j[2], None))
            customdata_d.extend((tx_vec[i], tx_vec[i + 1], None))
            colors1_d.extend(["orange"] * 3)

            # shear load on y and z axes
            p_i = p_start + qy_vec[i] * y_vec * scaling_q * scaling_global
            p_j = p_end + qy_vec[i + 1] * y_vec * scaling_q * scaling_global
            x1_b.extend((p_i[0], p_j[0], None))
            y1_b.extend((p_i[1], p_j[1], None))
            z1_b.extend((p_i[2], p_j[2], None))
            customdata_b.extend((qy_vec[i], qy_vec[i + 1], None))
            colors1_b.extend(["green"] * 3)
            p_i = p_start + qz_vec[i] * z_vec * scaling_q * scaling_global
            p_j = p_end + qz_vec[i + 1] * z_vec * scaling_q * scaling_global
            x1_b.extend((p_i[0], p_j[0], None))
            y1_b.extend((p_i[1], p_j[1], None))
            z1_b.extend((p_i[2], p_j[2], None))
            customdata_b.extend((qz_vec[i], qz_vec[i + 1], None))
            colors1_b.extend(["green"] * 3)

            # moment around z and y axes
            p_i = p_start - mz_vec[i] * y_vec * scaling_m * scaling_global
            p_j = p_end - mz_vec[i + 1] * y_vec * scaling_m * scaling_global
            # note: moments plotted upside down!
            x1_c.extend((p_i[0], p_j[0], None))
            y1_c.extend((p_i[1], p_j[1], None))
            z1_c.extend((p_i[2], p_j[2], None))
            customdata_c.extend((mz_vec[i], mz_vec[i + 1], None))
            colors1_c.extend(["blue"] * 3)
            # moment around y axis
            p_i = p_start - my_vec[i] * z_vec * scaling_m * scaling_global
            p_j = p_end - my_vec[i + 1] * z_vec * scaling_m * scaling_global
            x1_c.extend((p_i[0], p_j[0], None))
            y1_c.extend((p_i[1], p_j[1], None))
            z1_c.extend((p_i[2], p_j[2], None))
            customdata_c.extend((my_vec[i], my_vec[i + 1], None))
            colors1_c.extend(["blue"] * 3)

            # shear load combined
            p_i = (
                p_start
                + (qy_vec[i] * y_vec + qz_vec[i] * z_vec)
                * scaling_q
                * scaling_global
            )
            p_j = (
                p_end
                + (qy_vec[i + 1] * y_vec + qz_vec[i + 1] * z_vec)
                * scaling_q
                * scaling_global
            )
            x1_e.extend((p_i[0], p_j[0], None))
            y1_e.extend((p_i[1], p_j[1], None))
            z1_e.extend((p_i[2], p_j[2], None))
            customdata_e.extend(
                (
                    np.sqrt(qy_vec[i] ** 2 + qz_vec[i] ** 2),
                    np.sqrt(qy_vec[i + 1] ** 2 + qz_vec[i + 1] ** 2),
                    None,
                )
            )
            colors1_e.extend(["green"] * 3)

            # both moments combined!
            p_i = (
                p_start
                - mz_vec[i] * y_vec * scaling_m
                - my_vec[i] * z_vec * scaling_m * scaling_global
            )
            p_j = (
                p_end
                - mz_vec[i + 1] * y_vec * scaling_m
                - my_vec[i + 1] * z_vec * scaling_m * scaling_global
            )
            # note: moments plotted upside down!
            x1_f.extend((p_i[0], p_j[0], None))
            y1_f.extend((p_i[1], p_j[1], None))
            z1_f.extend((p_i[2], p_j[2], None))
            customdata_f.extend(
                (
                    np.sqrt(mz_vec[i] ** 2 + my_vec[i] ** 2),
                    np.sqrt(mz_vec[i + 1] ** 2 + my_vec[i + 1] ** 2),
                    None,
                )
            )
            colors1_f.extend(["blue"] * 3)

    dt_a = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_a,
            "y": y1_a,
            "z": z1_a,
            "visible": False,
            "customdata": customdata_a,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_a},
        }
    ]
    dt_b = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_b,
            "y": y1_b,
            "z": z1_b,
            "visible": False,
            "customdata": customdata_b,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_b},
        }
    ]
    dt_c = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_c,
            "y": y1_c,
            "z": z1_c,
            "visible": False,
            "customdata": customdata_c,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_c},
        }
    ]
    dt_d = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_d,
            "y": y1_d,
            "z": z1_d,
            "visible": False,
            "customdata": customdata_d,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_d},
        }
    ]
    dt_e = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_e,
            "y": y1_e,
            "z": z1_e,
            "visible": False,
            "customdata": customdata_e,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_e},
        }
    ]
    dt_f = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_f,
            "y": y1_f,
            "z": z1_f,
            "visible": False,
            "customdata": customdata_f,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_f},
        }
    ]

    fig_datastructure = dict(
        data=data_dict + dt_a + dt_b + dt_c + dt_d + dt_e + dt_f, layout=layout
    )
    fig = go.Figure(fig_datastructure)
    fig.update_layout(
        updatemenus=[
            dict(
                type="buttons",
                buttons=[
                    dict(
                        label="None",
                        method="update",
                        args=[
                            {"visible": [True] * len(data_dict) + [False] * 6}
                        ],
                    ),
                    dict(
                        label="Axial",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [True]
                                    + [False] * 5
                                )
                            }
                        ],
                    ),
                    dict(
                        label="Shear",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [False]
                                    + [True]
                                    + [False] * 4
                                )
                            }
                        ],
                    ),
                    dict(
                        label="Moment",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [False] * 2
                                    + [True]
                                    + [False] * 3
                                )
                            }
                        ],
                    ),
                    dict(
                        label="Torsion",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [False] * 3
                                    + [True]
                                    + [False] * 2
                                )
                            }
                        ],
                    ),
                    dict(
                        label="Shear (combined)",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [False] * 4
                                    + [True]
                                    + [False]
                                )
                            }
                        ],
                    ),
                    dict(
                        label="Moment (combined)",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [False] * 5
                                    + [True]
                                )
                            }
                        ],
                    ),
                ],
            )
        ]
    )

    if "pytest" not in sys.modules:
        if to_html_file:
            fig.write_html(to_html_file)
        else:
            fig.show()

    metadata = {
        "scaling_n": scaling_n,
        "scaling_q": scaling_q,
        "scaling_m": scaling_m,
        "scaling_t": scaling_t,
    }
    return metadata


def show_basic_forces_combo(
    combo,
    scaling_global,
    scaling_n,
    scaling_q,
    scaling_m,
    scaling_t,
    num_points,
    force_conversion=1.00,
    moment_conversion=1.00,
    global_axes=False,
    camera=None,
    subset_model=None,
    to_html_file=None,
):
    """
    Visualize the model and plot the enveloped frame element basic forces
    for a load combination.

    Arguments:
      combo: a load combination object
      step: the analysis step to be visualized
      scaling_global: I don't even remember what this
        does. It's kind of a mess right now.
      scaling_n:
      scaling_q:
      scaling_m:
      scaling_t:
      num_points: number of points to include in the basic force
        curves
      force_conversion: Conversion factor to be applied at the
        hover box data for forces (for unit conversions)
      moment_conversion: Conversion factor to be applied at the
        hover box data for moments (for unit conversions)
      global_axes: whether to show global axes
      camera: custom positioning of the camera
      subset_model: use this model instead of the one contained in the
        analysis object. It needs to be a subset of the original
        model. This can be used to only show the results for some part
        of a large model.
      to_html_file: If a path is specified, the figure is written in
        an html file instead of being shown.

    """

    # TODO: merge code repetitions with the previous function

    if subset_model:
        mdl = subset_model
    else:
        mdl = combo.mdl
    elms: list[
        Union[
            element.TrussBar, element.ElasticBeamColumn, element.DispBeamColumn
        ]
    ] = []
    elms.extend(mdl.list_of_specific_element(element.TrussBar))
    elms.extend(mdl.list_of_specific_element(element.ElasticBeamColumn))
    elms.extend(mdl.list_of_specific_element(element.DispBeamColumn))
    list_of_line_elements = [
        elm for elm in elms if not elm.visibility.skip_opensees_definition
    ]

    layout = graphics_common_3d.global_layout(mdl, camera)
    data_dict: list[dict[str, object]] = []

    # draw the frames
    add_data__frames_undeformed(data_dict, list_of_line_elements)
    # we also add axes so that we can see 2D plots
    ref_len = mdl.reference_length()
    if global_axes:
        add_data__global_axes(data_dict, ref_len)

    # Plot options:
    # a: axial
    # b: shear in local Y and Z
    # c: moment in Y and Z
    # d: torsion
    # e: shear combined
    # f: moment combined
    x1_a: list[Optional[float]] = []
    y1_a: list[Optional[float]] = []
    z1_a: list[Optional[float]] = []
    colors1_a: list[Optional[str]] = []
    customdata_a: list[Optional[float]] = []

    x1_b: list[Optional[float]] = []
    y1_b: list[Optional[float]] = []
    z1_b: list[Optional[float]] = []
    colors1_b: list[Optional[str]] = []
    customdata_b: list[Optional[float]] = []

    x1_c: list[Optional[float]] = []
    y1_c: list[Optional[float]] = []
    z1_c: list[Optional[float]] = []
    colors1_c: list[Optional[str]] = []
    customdata_c: list[Optional[float]] = []

    x1_d: list[Optional[float]] = []
    y1_d: list[Optional[float]] = []
    z1_d: list[Optional[float]] = []
    colors1_d: list[Optional[str]] = []
    customdata_d: list[Optional[float]] = []

    # (we do this to determine the internal forces
    #  for all elements before we even start plotting
    #  them, to be able to compute a nice scaling factor
    #  without having to then recalculate the basic forces)
    nx_vecs_min = {}
    qy_vecs_min = {}
    qz_vecs_min = {}
    tx_vecs_min = {}
    mz_vecs_min = {}
    my_vecs_min = {}
    nx_vecs_max = {}
    qy_vecs_max = {}
    qz_vecs_max = {}
    tx_vecs_max = {}
    mz_vecs_max = {}
    my_vecs_max = {}

    for elm in list_of_line_elements:
        if elm.visibility.hidden_basic_forces:
            continue

        df_min, df_max = combo.envelope_basic_forces(elm, num_points)

        # store results in the preallocated arrays

        nx_vecs_min[elm.uid] = df_min["nx"].to_numpy() * force_conversion
        qy_vecs_min[elm.uid] = df_min["qy"].to_numpy() * force_conversion
        qz_vecs_min[elm.uid] = df_min["qz"].to_numpy() * force_conversion
        tx_vecs_min[elm.uid] = df_min["tx"].to_numpy() * moment_conversion
        my_vecs_min[elm.uid] = df_min["my"].to_numpy() * moment_conversion
        mz_vecs_min[elm.uid] = df_min["mz"].to_numpy() * moment_conversion
        nx_vecs_max[elm.uid] = df_max["nx"].to_numpy() * force_conversion
        qy_vecs_max[elm.uid] = df_max["qy"].to_numpy() * force_conversion
        qz_vecs_max[elm.uid] = df_max["qz"].to_numpy() * force_conversion
        tx_vecs_max[elm.uid] = df_max["tx"].to_numpy() * moment_conversion
        my_vecs_max[elm.uid] = df_max["my"].to_numpy() * moment_conversion
        mz_vecs_max[elm.uid] = df_max["mz"].to_numpy() * moment_conversion

    # calculate scaling factors
    ref_len = mdl.reference_length()
    factor = 0.05
    nx_max = np.max(
        np.abs(
            np.column_stack(
                list(nx_vecs_min.values()) + list(nx_vecs_max.values())
            )
        )
    )
    scaling_n = force_scaling_factor(ref_len, nx_max, factor)
    if scaling_n > 1.0e8:
        scaling_t = 1.00
    qy_max = np.max(
        np.abs(
            np.column_stack(
                list(qy_vecs_min.values()) + list(qy_vecs_max.values())
            )
        )
    )
    qz_max = np.max(
        np.abs(
            np.column_stack(
                list(qz_vecs_min.values()) + list(qz_vecs_max.values())
            )
        )
    )
    scaling_qy = force_scaling_factor(ref_len, qy_max, factor)
    scaling_qz = force_scaling_factor(ref_len, qz_max, factor)
    if scaling_qy > 0.00 and scaling_qz > 0.00:
        scaling_q = np.minimum(scaling_qy, scaling_qz)
    elif scaling_qy == 0.00:
        scaling_q = scaling_qz
    elif scaling_qz == 0.00:
        scaling_q = scaling_qy
    else:
        scaling_q = 0.00
    if scaling_q > 1.0e8:
        scaling_q = 1.00
    my_max = np.max(
        np.abs(
            np.column_stack(
                list(my_vecs_min.values()) + list(my_vecs_max.values())
            )
        )
    )
    mz_max = np.max(
        np.abs(
            np.column_stack(
                list(mz_vecs_min.values()) + list(mz_vecs_max.values())
            )
        )
    )
    scaling_my = force_scaling_factor(ref_len, my_max, factor)
    scaling_mz = force_scaling_factor(ref_len, mz_max, factor)
    if scaling_my > 0.00 and scaling_mz > 0.00:
        scaling_m = np.minimum(scaling_my, scaling_mz)
    elif scaling_my == 0.00:
        scaling_m = scaling_mz
    elif scaling_mz == 0.00:
        scaling_m = scaling_my
    else:
        scaling_m = 0.00
    if scaling_m > 1.0e8:
        scaling_m = 1.00

    for elm in list_of_line_elements:
        if elm.visibility.hidden_basic_forces:
            continue

        # retrieve results from the preallocated arrays
        nx_vec_min = nx_vecs_min[elm.uid]
        qy_vec_min = qy_vecs_min[elm.uid]
        qz_vec_min = qz_vecs_min[elm.uid]
        tx_vec_min = tx_vecs_min[elm.uid]
        my_vec_min = my_vecs_min[elm.uid]
        mz_vec_min = mz_vecs_min[elm.uid]

        nx_vec_max = nx_vecs_max[elm.uid]
        qy_vec_max = qy_vecs_max[elm.uid]
        qz_vec_max = qz_vecs_max[elm.uid]
        tx_vec_max = tx_vecs_max[elm.uid]
        my_vec_max = my_vecs_max[elm.uid]
        mz_vec_max = mz_vecs_max[elm.uid]

        if isinstance(
            elm, (element.ElasticBeamColumn, element.DispBeamColumn)
        ):
            x_vec = elm.geomtransf.x_axis
            y_vec = elm.geomtransf.y_axis
            z_vec = elm.geomtransf.z_axis
            i_pos = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
            p_i = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
            p_j = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
        else:
            p_i = np.array(elm.nodes[0].coords)
            p_j = np.array(elm.nodes[1].coords)
            i_pos = np.array(elm.nodes[0].coords)
            (
                x_vec,
                y_vec,
                z_vec,
            ) = transformations.local_axes_from_points_and_angle(
                p_i, p_j, 0.00
            )

        len_clr = np.linalg.norm(p_i - p_j)
        t_vec = np.linspace(0.00, len_clr, num=num_points)

        for i in range(num_points - 1):
            p_start = i_pos + t_vec[i] * x_vec
            p_end = i_pos + t_vec[i + 1] * x_vec

            # axial load
            p_i = p_start + nx_vec_min[i] * y_vec * scaling_n * scaling_global
            p_j = (
                p_end + nx_vec_min[i + 1] * y_vec * scaling_n * scaling_global
            )
            x1_a.extend((p_i[0], p_j[0], None))
            y1_a.extend((p_i[1], p_j[1], None))
            z1_a.extend((p_i[2], p_j[2], None))
            customdata_a.extend((nx_vec_min[i], nx_vec_min[i + 1], None))
            colors1_a.extend(["red"] * 3)
            p_i = p_start + nx_vec_max[i] * y_vec * scaling_n * scaling_global
            p_j = (
                p_end + nx_vec_max[i + 1] * y_vec * scaling_n * scaling_global
            )
            x1_a.extend((p_i[0], p_j[0], None))
            y1_a.extend((p_i[1], p_j[1], None))
            z1_a.extend((p_i[2], p_j[2], None))
            customdata_a.extend((nx_vec_max[i], nx_vec_max[i + 1], None))
            colors1_a.extend(["green"] * 3)

            # torsion
            p_i = p_start + tx_vec_min[i] * z_vec * scaling_t * scaling_global
            p_j = (
                p_end + tx_vec_min[i + 1] * z_vec * scaling_t * scaling_global
            )
            x1_d.extend((p_i[0], p_j[0], None))
            y1_d.extend((p_i[1], p_j[1], None))
            z1_d.extend((p_i[2], p_j[2], None))
            customdata_d.extend((tx_vec_min[i], tx_vec_min[i + 1], None))
            colors1_d.extend(["red"] * 3)
            p_i = p_start + tx_vec_max[i] * z_vec * scaling_t * scaling_global
            p_j = (
                p_end + tx_vec_max[i + 1] * z_vec * scaling_t * scaling_global
            )
            x1_d.extend((p_i[0], p_j[0], None))
            y1_d.extend((p_i[1], p_j[1], None))
            z1_d.extend((p_i[2], p_j[2], None))
            customdata_d.extend((tx_vec_max[i], tx_vec_max[i + 1], None))
            colors1_d.extend(["green"] * 3)

            # shear load on y and z axes
            p_i = p_start + qy_vec_min[i] * y_vec * scaling_q * scaling_global
            p_j = (
                p_end + qy_vec_min[i + 1] * y_vec * scaling_q * scaling_global
            )
            x1_b.extend((p_i[0], p_j[0], None))
            y1_b.extend((p_i[1], p_j[1], None))
            z1_b.extend((p_i[2], p_j[2], None))
            customdata_b.extend((qy_vec_min[i], qy_vec_min[i + 1], None))
            colors1_b.extend(["red"] * 3)
            p_i = p_start + qy_vec_max[i] * y_vec * scaling_q * scaling_global
            p_j = (
                p_end + qy_vec_max[i + 1] * y_vec * scaling_q * scaling_global
            )
            x1_b.extend((p_i[0], p_j[0], None))
            y1_b.extend((p_i[1], p_j[1], None))
            z1_b.extend((p_i[2], p_j[2], None))
            customdata_b.extend((qy_vec_max[i], qy_vec_max[i + 1], None))
            colors1_b.extend(["green"] * 3)
            p_i = p_start + qz_vec_min[i] * z_vec * scaling_q * scaling_global
            p_j = (
                p_end + qz_vec_min[i + 1] * z_vec * scaling_q * scaling_global
            )
            x1_b.extend((p_i[0], p_j[0], None))
            y1_b.extend((p_i[1], p_j[1], None))
            z1_b.extend((p_i[2], p_j[2], None))
            customdata_b.extend((qz_vec_min[i], qz_vec_min[i + 1], None))
            colors1_b.extend(["red"] * 3)
            p_i = p_start + qz_vec_max[i] * z_vec * scaling_q * scaling_global
            p_j = (
                p_end + qz_vec_max[i + 1] * z_vec * scaling_q * scaling_global
            )
            x1_b.extend((p_i[0], p_j[0], None))
            y1_b.extend((p_i[1], p_j[1], None))
            z1_b.extend((p_i[2], p_j[2], None))
            customdata_b.extend((qz_vec_max[i], qz_vec_max[i + 1], None))
            colors1_b.extend(["green"] * 3)

            # moment around z and y axes
            p_i = p_start - mz_vec_min[i] * y_vec * scaling_m * scaling_global
            p_j = (
                p_end - mz_vec_min[i + 1] * y_vec * scaling_m * scaling_global
            )
            # note: moments plotted upside down!
            x1_c.extend((p_i[0], p_j[0], None))
            y1_c.extend((p_i[1], p_j[1], None))
            z1_c.extend((p_i[2], p_j[2], None))
            customdata_c.extend((mz_vec_min[i], mz_vec_min[i + 1], None))
            colors1_c.extend(["red"] * 3)
            p_i = p_start - mz_vec_max[i] * y_vec * scaling_m * scaling_global
            p_j = (
                p_end - mz_vec_max[i + 1] * y_vec * scaling_m * scaling_global
            )
            # note: moments plotted upside down!
            x1_c.extend((p_i[0], p_j[0], None))
            y1_c.extend((p_i[1], p_j[1], None))
            z1_c.extend((p_i[2], p_j[2], None))
            customdata_c.extend((mz_vec_max[i], mz_vec_max[i + 1], None))
            colors1_c.extend(["green"] * 3)
            # moment around y axis
            p_i = p_start - my_vec_min[i] * z_vec * scaling_m * scaling_global
            p_j = (
                p_end - my_vec_min[i + 1] * z_vec * scaling_m * scaling_global
            )
            x1_c.extend((p_i[0], p_j[0], None))
            y1_c.extend((p_i[1], p_j[1], None))
            z1_c.extend((p_i[2], p_j[2], None))
            customdata_c.extend((my_vec_min[i], my_vec_min[i + 1], None))
            colors1_c.extend(["red"] * 3)
            p_i = p_start - my_vec_max[i] * z_vec * scaling_m * scaling_global
            p_j = (
                p_end - my_vec_max[i + 1] * z_vec * scaling_m * scaling_global
            )
            x1_c.extend((p_i[0], p_j[0], None))
            y1_c.extend((p_i[1], p_j[1], None))
            z1_c.extend((p_i[2], p_j[2], None))
            customdata_c.extend((my_vec_max[i], my_vec_max[i + 1], None))
            colors1_c.extend(["green"] * 3)

    dt_a = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_a,
            "y": y1_a,
            "z": z1_a,
            "visible": False,
            "customdata": customdata_a,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_a},
        }
    ]
    dt_b = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_b,
            "y": y1_b,
            "z": z1_b,
            "visible": False,
            "customdata": customdata_b,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_b},
        }
    ]
    dt_c = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_c,
            "y": y1_c,
            "z": z1_c,
            "visible": False,
            "customdata": customdata_c,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_c},
        }
    ]
    dt_d = [
        {
            "type": "scatter3d",
            "mode": "lines",
            "x": x1_d,
            "y": y1_d,
            "z": z1_d,
            "visible": False,
            "customdata": customdata_d,
            "hovertemplate": " %{customdata:.0f}<br><extra></extra>",
            "line": {"width": 3, "color": colors1_d},
        }
    ]

    fig_datastructure = dict(
        data=data_dict + dt_a + dt_b + dt_c + dt_d, layout=layout
    )
    fig = go.Figure(fig_datastructure)
    fig.update_layout(
        updatemenus=[
            dict(
                type="buttons",
                buttons=[
                    dict(
                        label="None",
                        method="update",
                        args=[
                            {"visible": [True] * len(data_dict) + [False] * 4}
                        ],
                    ),
                    dict(
                        label="Axial",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [True]
                                    + [False] * 3
                                )
                            }
                        ],
                    ),
                    dict(
                        label="Shear",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [False]
                                    + [True]
                                    + [False] * 2
                                )
                            }
                        ],
                    ),
                    dict(
                        label="Moment",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [False] * 2
                                    + [True]
                                    + [False]
                                )
                            }
                        ],
                    ),
                    dict(
                        label="Torsion",
                        method="update",
                        args=[
                            {
                                "visible": (
                                    [True] * len(data_dict)
                                    + [False] * 3
                                    + [True]
                                )
                            }
                        ],
                    ),
                ],
            )
        ]
    )

    if "pytest" not in sys.modules:
        if to_html_file:
            fig.write_html(to_html_file)
        else:
            fig.show()

    metadata = {
        "scaling_n": scaling_n,
        "scaling_q": scaling_q,
        "scaling_m": scaling_m,
        "scaling_t": scaling_t,
    }
    return metadata



"""
Common definitions
"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import re
from pprint import pprint
from typing import OrderedDict
from typing import Optional
from typing import Any

# very big, very small numbers used for
# comparing floats and hashing
EPSILON = 1.00e-6
ALPHA = 10000000.00

# gravitational acceleration
G_CONST_IMPERIAL = 386.22  # in/s**2
G_CONST_SI = 9.81  # m/s**2

# quantities to use for extreme stiffnesses
STIFF_ROT = 1.0e15
STIFF = 1.0e10
TINY = 1.0e-12


def methods(obj: object) -> list[str]:
    """
    Returns the names of all methods of an object,
    excluding the dunder methods.

    Example:
        >>> class TestClass:
        ...     def method_1(self):
        ...         pass
        ...
        ...     def method_2(self):
        ...         pass
        ...
        >>> obj = TestClass()
        >>> methods(obj)
        ['method_1', 'method_2']
    """
    object_methods = [
        method_name
        for method_name in dir(obj)
        if callable(getattr(obj, method_name))
    ]
    pattern = r"__.*__"
    return [s for s in object_methods if not re.match(pattern, s)]


def print_methods(obj: object) -> None:
    """
    Prints the methods of an object.
    """
    object_methods = methods(obj)
    pprint(object_methods)


def print_dir(obj: object) -> None:
    """
    Prints the entire output of `dir()` of an object
    """
    pprint(dir(obj))


def previous_element(dct: OrderedDict[Any, Any], key: Any) -> Optional[Any]:
    """
    Returns the value of the element that comes before the given key
    in an ordered dictionary.
    If the key is not in the dictionary, or if it is the first element
    in the dictionary, returns None.

    Arguments:
        dct: An ordered dictionary.
        key: The key of the element whose previous element we want to
        find.

    Returns:
        The value of the element that comes before the given key in
        the dictionary, or None if there is no such element.

    Example:
        >>> dct = OrderedDict([(1, 'a'), (2, 'b'), (3, 'c')])
        >>> previous_element(dct, 2)  # Returns 'a'
        'a'
        >>> previous_element(dct, 3)  # Returns 'b'
        'b'
        >>> previous_element(dct, 1)  # Returns None

        >>> previous_element(dct, 4)  # Returns None

    """
    if key in dct:
        key_list = list(dct.keys())
        idx = key_list.index(key)
        if idx == 0:
            ans = None
        else:
            ans = dct[key_list[idx - 1]]
    else:
        ans = None
    return ans



"""
Objects that generate ZeroLength elements.

"""

#
#   _|_|      _|_|_|  _|      _|    _|_|_|
# _|    _|  _|        _|_|  _|_|  _|
# _|    _|    _|_|    _|  _|  _|  _|  _|_|
# _|    _|        _|  _|      _|  _|    _|
#   _|_|    _|_|_|    _|      _|    _|_|_|
#
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

# pylint: disable=unused-argument


from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Optional
import numpy as np
from ..ops.section import ElasticSection
from ..ops.uniaxial_material import UniaxialMaterial
from ..ops.uniaxial_material import Elastic
from ..ops.uniaxial_material import Steel02
from ..ops.uniaxial_material import Pinching4
from ..ops.uniaxial_material import Hysteretic
from .material_gen import MaterialGenerator

if TYPE_CHECKING:
    from ..model import Model
    from ..physical_material import PhysicalMaterial


def fix_all(
    model: Model, **kwargs: dict[object, object]
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    Fixed in all directions.

    """

    dirs = [1, 2, 3, 4, 5, 6]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat] * 3
    return dirs, mats


def release_6(
    model: Model, **kwargs: dict[object, object]
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    Frees strong axis bending.

    """

    dirs = [1, 2, 3, 4, 5]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat] * 2
    return dirs, mats


def release_5(
    model: Model, **kwargs: dict[object, object]
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    Frees weak axis bending.

    """

    dirs = [1, 2, 3, 4, 6]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat] * 2
    return dirs, mats


def release_56(
    model: Model, **kwargs: dict[object, object]
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    Frees both strong and weak axis bending.

    """

    dirs = [1, 2, 3, 4]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat]
    return dirs, mats


def imk_6(
    model: Model,
    element_length: float,
    lboverl: float,
    loverh: float,
    rbs_factor: Optional[float],
    consider_composite: bool,
    axial_load_ratio: float,
    section: ElasticSection,
    physical_material: PhysicalMaterial,
    **kwargs: dict[object, object],
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    Lignos, D. G., & Krawinkler, H. (2011). Deterioration modeling of
    steel components in support of collapse prediction of steel moment
    frames under earthquake loading. Journal of Structural
    Engineering-Reston, 137(11), 1291.

    Elkady, A., & Lignos, D. G. (2014). Modeling of the composite
    action in fully restrained beam‐to‐column connections:
    implications in the seismic design and collapse capacity of steel
    special moment frames. Earthquake Engineering & Structural
    Dynamics, 43(13), 1935-1954.

    """

    moment_modifier = kwargs.get('moment_modifier', 1.00)
    n_parameter = kwargs.get('n_parameter', 0.00)
    only_elastic = kwargs.get('only_elastic', False)

    mat_generator = MaterialGenerator(model)
    mat = mat_generator.generate_steel_w_imk_material(
        section,
        physical_material,
        element_length,
        lboverl,
        loverh,
        rbs_factor,
        consider_composite,
        axial_load_ratio,
        direction="strong",
        moment_modifier=moment_modifier,
        n_parameter=n_parameter,
        only_elastic=only_elastic,
    )
    dirs = [1, 2, 3, 4, 5, 6]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat] * 2 + [mat]
    return dirs, mats


def imk_56(
    model: Model,
    element_length: float,
    lboverl: float,
    loverh: float,
    rbs_factor: Optional[float],
    consider_composite: bool,
    axial_load_ratio: float,
    section: ElasticSection,
    physical_material: PhysicalMaterial,
    **kwargs: dict[object, object],
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    release in the weak axis bending direction,
    :func:`~osmg.gen.zerolength_gen.imk_6` in the strong axis bending
    direction

    """

    moment_modifier = kwargs.get('moment_modifier', 1.00)
    n_parameter = kwargs.get('n_parameter', 0.00)
    only_elastic = kwargs.get('only_elastic', False)

    mat_generator = MaterialGenerator(model)
    mat_strong = mat_generator.generate_steel_w_imk_material(
        section,
        physical_material,
        element_length,
        lboverl,
        loverh,
        rbs_factor,
        consider_composite,
        axial_load_ratio,
        direction="strong",
        moment_modifier=moment_modifier,
        n_parameter=n_parameter,
        only_elastic=only_elastic,
    )
    mat_weak = mat_generator.generate_steel_w_imk_material(
        section,
        physical_material,
        element_length,
        lboverl,
        loverh,
        rbs_factor,
        consider_composite,
        axial_load_ratio,
        direction="weak",
        moment_modifier=moment_modifier,
        only_elastic=only_elastic,
    )
    dirs = [1, 2, 3, 4, 5, 6]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat, mat_weak, mat_strong]
    return dirs, mats


def imk_6_release_5(
    model: Model,
    element_length: float,
    lboverl: float,
    loverh: float,
    rbs_factor: Optional[float],
    consider_composite: bool,
    axial_load_ratio: float,
    section: ElasticSection,
    physical_material: PhysicalMaterial,
    **kwargs: dict[object, object],
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    release in the weak axis bending direction,
    :func:`~osmg.gen.zerolength_gen.imk_6` in the strong axis bending
    direction

    """

    moment_modifier = kwargs.get('moment_modifier', 1.00)
    n_parameter = kwargs.get('n_parameter', 0.00)
    only_elastic = kwargs.get('only_elastic', False)

    mat_generator = MaterialGenerator(model)
    mat_strong = mat_generator.generate_steel_w_imk_material(
        section,
        physical_material,
        element_length,
        lboverl,
        loverh,
        rbs_factor,
        consider_composite,
        axial_load_ratio,
        direction="strong",
        moment_modifier=moment_modifier,
        n_parameter=n_parameter,
        only_elastic=only_elastic,
    )
    dirs = [1, 2, 3, 4, 6]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat, mat_strong]
    return dirs, mats


def gravity_shear_tab(
    model: Model,
    consider_composite: bool,
    section: ElasticSection,
    physical_material: PhysicalMaterial,
    **kwargs: dict[object, object],
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    Elkady, A., & Lignos, D. G. (2015). Effect of gravity framing on
    the overstrength and collapse capacity of steel frame buildings
    with perimeter special moment frames. Earthquake Engineering &
    Structural Dynamics, 44(8), 1289-1307.

    """

    assert section.name[0] == "W", "Error: Only W sections can be used."
    assert isinstance(section, ElasticSection)
    assert (
        model.settings.imperial_units
    ), "Error: Only imperial units supported."
    assert section.properties

    moment_modifier = kwargs.get('moment_modifier', 1.00)

    # Yield stress
    assert isinstance(moment_modifier, float)
    mat_fy = physical_material.f_y / 1.0e3
    # Plastic modulus (unreduced)
    sec_zx = section.properties["Zx"]
    # Plastic moment of the section
    sec_mp = sec_zx * mat_fy * 1.0e3 * moment_modifier

    if not consider_composite:
        m_max_pos = 0.121 * sec_mp
        m_max_neg = 0.121 * sec_mp
        m1_p = +0.521 * m_max_pos
        m1_n = -0.521 * m_max_neg
        m2_p = +0.967 * m_max_pos
        m2_n = -0.967 * m_max_neg
        m3_p = +1.000 * m_max_pos
        m3_n = -1.000 * m_max_pos
        m4_p = +0.901 * m_max_pos
        m4_n = -0.901 * m_max_neg
        th_1_p = 0.0045
        th_1_n = -0.0045
        th_2_p = 0.0465
        th_2_n = -0.0465
        th_3_p = 0.0750
        th_3_n = -0.0750
        th_4_p = 0.1000
        th_4_n = -0.1000
        rdispp = 0.57
        rdispn = 0.57
        rforcep = 0.40
        rforcen = 0.40
        uforcep = 0.05
        uforcen = 0.05
        gklim = 0.2
        gdlim = 0.1
        gflim = 0.0
        g_e = 10
        dmgtype = "energy"
    else:
        m_max_pos = 0.35 * sec_mp
        m_max_neg = 0.64 * 0.35 * sec_mp
        m1_p = +0.250 * m_max_pos
        m1_n = -0.250 * m_max_neg
        m2_p = +1.000 * m_max_pos
        m2_n = -1.000 * m_max_neg
        m3_p = +1.001 * m_max_pos
        m3_n = -1.001 * m_max_pos
        m4_p = +0.530 * m_max_pos
        m4_n = -0.540 * m_max_neg
        th_1_p = 0.0042
        th_1_n = -0.0042
        th_2_p = 0.0200
        th_2_n = -0.0110
        th_3_p = 0.0390
        th_3_n = -0.0300
        th_4_p = 0.0400
        th_4_n = -0.0550
        rdispp = 0.40
        rdispn = 0.50
        rforcep = 0.13
        rforcen = 0.53
        uforcep = 0.01
        uforcen = 0.05
        gklim = 0.30
        gdlim = 0.05
        gflim = 0.05
        g_e = 10
        dmgtype = "energy"

    mat = Pinching4(
        model.uid_generator.new("uniaxial material"),
        "auto_gravity_shear_tab",
        m1_p,
        th_1_p,
        m2_p,
        th_2_p,
        m3_p,
        th_3_p,
        m4_p,
        th_4_p,
        m1_n,
        th_1_n,
        m2_n,
        th_2_n,
        m3_n,
        th_3_n,
        m4_n,
        th_4_n,
        rdispp,
        rforcep,
        uforcep,
        rdispn,
        rforcen,
        uforcen,
        0.00,
        0.00,
        0.00,
        0.00,
        gklim,
        0.00,
        0.00,
        0.00,
        0.00,
        gdlim,
        0.00,
        0.00,
        0.00,
        0.00,
        gflim,
        g_e,
        dmgtype,
    )
    dirs = [1, 2, 3, 4, 5, 6]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    release_mat = mat_repo.retrieve_by_attr("name", "release")
    mats = [fix_mat] * 3 + [fix_rot_mat, release_mat, mat]
    return dirs, mats


def steel_w_col_pz(
    model: Model,
    section: ElasticSection,
    physical_material: PhysicalMaterial,
    pz_length: float,
    pz_doubler_plate_thickness: float,
    pz_hardening: float,
    only_elastic: bool = False,
    moment_modifier: float = 1.00,
    **kwargs: dict[object, object],
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    Gupta, A., & Krawinkler, H. (1999). Seismic demands for the
    performance evaluation of steel moment resisting frame
    structures. Rep. No. 132.

    """

    assert section.name[0] == "W", "Error: Only W sections can be used."
    assert isinstance(section, ElasticSection)
    assert (
        model.settings.imperial_units
    ), "Error: Only imperial units supported."
    assert section.properties
    f_y = physical_material.f_y
    hardening = pz_hardening
    d_c = section.properties["d"]
    bfc = section.properties["bf"]
    t_p = section.properties["tw"] + pz_doubler_plate_thickness
    t_f = section.properties["tf"]
    v_y = 0.55 * f_y * d_c * t_p
    g_mod = physical_material.g_mod
    k_e = 0.95 * g_mod * t_p * d_c
    k_p = 0.95 * g_mod * bfc * t_f**2 / pz_length
    gamma_1 = v_y / k_e
    gamma_2 = 4.0 * gamma_1
    gamma_3 = 100.0 * gamma_1
    m1y = gamma_1 * k_e * pz_length * moment_modifier
    m2y = m1y + k_p * pz_length * (gamma_2 - gamma_1) * moment_modifier
    m3y = (
        m2y
        + (hardening * k_e * pz_length) * (gamma_3 - gamma_2) * moment_modifier
    )

    if only_elastic:
        mat: UniaxialMaterial = Elastic(
            model.uid_generator.new("uniaxial material"),
            "auto_steel_W_PZ",
            m1y / gamma_1,
        )
    else:
        mat = Hysteretic(
            model.uid_generator.new("uniaxial material"),
            "auto_steel_W_PZ",
            (m1y, gamma_1),
            (m2y, gamma_2),
            (m3y, gamma_3),
            (-m1y, -gamma_1),
            (-m2y, -gamma_2),
            (-m3y, -gamma_3),
            0.25,
            0.75,
            0.00,
            0.00,
            0.00,
        )
    dirs = [1, 2, 3, 4, 5, 6]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat] * 2 + [mat]
    return dirs, mats


def steel_w_col_pz_updated(
    model: Model,
    section: ElasticSection,
    physical_material: PhysicalMaterial,
    pz_length: float,
    pz_doubler_plate_thickness: float,
    axial_load_ratio: float,
    slab_depth: float,
    consider_composite: bool,
    location: str,
    only_elastic: bool = False,
    moment_modifier: float = 1.00,
    **kwargs: dict[object, object],
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    Skiadopoulos, A., Elkady, A. and D. G. Lignos (2020). "Proposed
    Panel Zone Model for Seismic Design of Steel Moment-Resisting
    Frames." ASCE Journal of Structural Engineering. DOI:
    10.1061/(ASCE)ST.1943-541X.0002935.

    """

    assert section.name[0] == "W", "Error: Only W sections can be used."
    assert isinstance(section, ElasticSection)
    assert (
        model.settings.imperial_units
    ), "Error: Only imperial units supported."
    assert section.properties
    f_y = physical_material.f_y
    e_mod = physical_material.e_mod
    g_mod = physical_material.g_mod
    tw_Col = section.properties["tw"]
    tdp = pz_doubler_plate_thickness
    d_Col = section.properties["d"]
    d_Beam = pz_length
    tf_Col = section.properties["tf"]
    bf_Col = section.properties["bf"]
    Ix_Col = section.properties["Ix"]
    ts = slab_depth
    n = axial_load_ratio
    trib = slab_depth

    tpz = tw_Col + tdp  # total PZ thickness
    # effective depth in positive moment
    d_BeamP = d_Beam
    if consider_composite:
        d_BeamP = d_Beam + trib + 0.50 * ts
    # effective depth in negative moment
    d_BeamN = d_Beam

    # Stiffness Calculation
    Ks = tpz * (d_Col - tf_Col) * g_mod
    Kb = (
        12.0
        * e_mod
        * (Ix_Col + tdp * ((d_Col - 2.0 * tf_Col) ** 3) / 12.00)
        / (d_Beam**3)
        * d_Beam
    )
    Ke = Ks * Kb / (Ks + Kb)

    # flange stiffness: shear contribution
    Ksf = 2.0 * (bf_Col * tf_Col) * g_mod
    # flange stiffness: bending contribution
    Kbf = (
        2.0
        * 12.0
        * e_mod
        * bf_Col
        * (tf_Col**3)
        / 12.0
        / (d_Beam**3)
        * d_Beam
    )
    # flange stiffness: total contribution
    Kef = (Ksf * Kbf) / (Ksf + Kbf)

    ay = (0.58 * Kef / Ke + 0.88) / (1.0 - Kef / Ke)

    aw_eff_4gamma = 1.10
    aw_eff_6gamma = 1.15

    af_eff_4gamma = 0.93 * Kef / Ke + 0.015
    af_eff_6gamma = 1.05 * Kef / Ke + 0.020
    # reduction factor accounting for axial load
    r = np.sqrt(1.0 - (n**2))

    Vy = r * 0.577 * f_y * ay * (d_Col - tf_Col) * tpz
    # Plastic Shear Force at 4 gammaY
    Vp_4gamma = (
        r
        * 0.577
        * f_y
        * (
            aw_eff_4gamma * (d_Col - tf_Col) * tpz
            + af_eff_4gamma * (bf_Col - tw_Col) * 2 * tf_Col
        )
    )
    # Plastic Shear Force at 6 gammaY
    Vp_6gamma = (
        r
        * 0.577
        * f_y
        * (
            aw_eff_6gamma * (d_Col - tf_Col) * tpz
            + af_eff_6gamma * (bf_Col - tw_Col) * 2 * tf_Col
        )
    )

    gamma_y = Vy / Ke
    gamma4_y = 4.0 * gamma_y

    My_P = Vy * d_BeamP
    Mp_4gamma_P = Vp_4gamma * d_BeamP
    Mp_6gamma_P = Vp_6gamma * d_BeamP

    My_N = Vy * d_BeamN
    Mp_4gamma_N = Vp_4gamma * d_BeamN
    Mp_6gamma_N = Vp_6gamma * d_BeamN

    Slope_4to6gamma_y_P = (Mp_6gamma_P - Mp_4gamma_P) / (2.0 * gamma_y)
    Slope_4to6gamma_y_N = (Mp_6gamma_N - Mp_4gamma_N) / (2.0 * gamma_y)

    # Defining the 3 Points used to construct the trilinear backbone curve
    gamma1 = gamma_y
    gamma2 = gamma4_y
    gamma3 = 100.0 * gamma_y

    M1_P = My_P
    M2_P = Mp_4gamma_P
    M3_P = Mp_4gamma_P + Slope_4to6gamma_y_P * (100 * gamma_y - gamma4_y)

    M1_N = My_N
    M2_N = Mp_4gamma_N
    M3_N = Mp_4gamma_N + Slope_4to6gamma_y_N * (100 * gamma_y - gamma4_y)

    # apply moment modifier
    M1_P *= moment_modifier
    M2_P *= moment_modifier
    M3_P *= moment_modifier
    M1_N *= moment_modifier
    M2_N *= moment_modifier
    M3_N *= moment_modifier

    if not consider_composite:
        args = (
            (M1_N, gamma1),
            (M2_N, gamma2),
            (M3_N, gamma3),
            (-M1_N, -gamma1),
            (-M2_N, -gamma2),
            (-M3_N, -gamma3),
            0.25,
            0.75,
            0.0,
            0.0,
            0.0,
        )
    elif location == 'interior':
        args = (
            (M1_P, gamma1),
            (M2_P, gamma2),
            (M3_P, gamma3),
            (-M1_P, -gamma1),
            (-M2_P, -gamma2),
            (-M3_P, -gamma3),
            0.25,
            0.75,
            0.0,
            0.0,
            0.0,
        )
    elif location == 'exterior_first':
        args = (
            (M1_N, gamma1),
            (M2_N, gamma2),
            (M3_N, gamma3),
            (-M1_P, -gamma1),
            (-M2_P, -gamma2),
            (-M3_P, -gamma3),
            0.25,
            0.75,
            0.0,
            0.0,
            0.0,
        )
    elif location == 'exterior_last':
        args = (
            (M1_P, gamma1),
            (M2_P, gamma2),
            (M3_P, gamma3),
            (-M1_N, -gamma1),
            (-M2_N, -gamma2),
            (-M3_N, -gamma3),
            0.25,
            0.75,
            0.0,
            0.0,
            0.0,
        )
    else:
        raise ValueError(f'Invalid Location: {location}')

    if only_elastic:
        m1y, gamma_1 = args[0]  # type: ignore
        mat: UniaxialMaterial = Elastic(
            model.uid_generator.new("uniaxial material"),
            "auto_steel_W_PZ",
            m1y / gamma_1,  # type: ignore
        )
    else:
        mat = Hysteretic(
            model.uid_generator.new("uniaxial material"),
            "auto_steel_W_pz_updated",
            *args,
        )
    # minmaxmat = MinMax(
    #     model.uid_generator.new("uniaxial material"),
    #     "auto_steel_W_pz_updated_minmax",
    #     mat, gammaU_N, gammaU_P)
    # print(mat.ops_args())
    dirs = [1, 2, 3, 4, 5, 6]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat] * 2 + [mat]
    return dirs, mats


def steel_brace_gusset(
    model: Model,
    physical_mat: PhysicalMaterial,
    d_brace: float,
    l_c: float,
    t_p: float,
    l_b: float,
    **kwargs: dict[object, object],
) -> tuple[list[int], list[UniaxialMaterial]]:
    """
    Hsiao, P-C., Lehman, D.E., and Roeder, C.W., 2012, Improved
    analytical model for special concentrically braced frames, Journal
    of Constructional Steel Research, Vol. 73, pp 80-94.

    Arguments:
      model: Model object
      physical_mat: physical material object
      d_brace: brace section height
      l_c: brace-to-gusset connection length
      t_p: gusset plate thickness
      l_b: gusset plate average buckling length

    """

    var_w = d_brace + 2.00 * l_c * np.tan(30.00 / 180.00 * np.pi)
    var_i = var_w * t_p**3 / 12.00
    var_z = var_w * t_p**2 / 6.00
    f_y = physical_mat.f_y
    var_e = physical_mat.e_mod
    var_g = physical_mat.g_mod
    var_my = var_z * f_y
    var_k_rot = var_e * var_i / l_b
    var_b = 0.01
    gusset_mat = Steel02(
        model.uid_generator.new("uniaxial material"),
        "auto_steel_gusset",
        var_my,
        var_k_rot,
        var_g,
        var_b,
        20.00,
        0.925,
        0.15,
        0.0005,
        0.014,
        0.0005,
        0.01,
        0.00,
    )
    dirs = [1, 2, 3, 4, 5, 6]
    mat_repo = model.uniaxial_materials
    fix_mat = mat_repo.retrieve_by_attr("name", "fix")
    fix_rot_mat = mat_repo.retrieve_by_attr("name", "fix_rot")
    mats = [fix_mat] * 3 + [fix_rot_mat, gusset_mat, fix_mat]
    return dirs, mats
